#include <alloca.h>
#include <math.h>
#include <omp.h>
#include <pthread.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

// #include "../contrib/orlp/ipow.h"
#define STB_DS_IMPLEMENTATION
#define STBDS_NO_SHORT_NAMES
#include "../contrib/stb/stb_ds.h"
#include "../includes/cg.h"
#include "../includes/matrix.h"
#include "../includes/permutation.h"

static inline _Bool IsOdd(size_t value) { return (_Bool)(value & 1); }

static _Bool ReadArguments(char *from, size_t *into, const char *varname) {
  if (!sscanf(from, "%zu", into)) {
    fprintf(stderr, "Failed to parse %s as a positive integer.\n", varname);
    return 0;
  }
  return 1;
}

static inline void AdvanceConditionalPermutations(
    permutation_generator_t *generators, const size_t count) {
  PermutationNext(generators + 0);
  for (size_t i = 0; i < count - 1; i++) {
    if (generators[i].exhausted) {
      PermutationReset(generators + i);
      PermutationNext(generators + i + 1);
    } else {
      break;
    }
  }
  // If generators[-1] is exhausted, the conditional permutations are exhausted.
}

static inline void ResetConditionalPermutations(
    permutation_generator_t *generators, const size_t count) {
  for (size_t i = 0; i < count; i++) {
    PermutationReset(generators + i);
  }
}

typedef enum {
  kRowFmt,
  kCgFmt,
  kPFmt,
  kPRawFmt,
} arg_format_t;

typedef struct {
  size_t a_out;
  size_t b_out;
  size_t a_in;
  size_t b_in;
  size_t bin_size;
  arg_format_t formatting;
} args_t;

static args_t ParseArgs(size_t argc, char *argv[]) {
  char **positionals = NULL;
  char **flags = NULL;

  for (size_t i = 1; i < argc; i++) {
    char *arg = argv[i];

    if (*arg == '-' && *(arg + 1) == '-') {
      stbds_arrput(flags, arg + 2);
    } else {
      stbds_arrput(positionals, arg);
    }
  }

  size_t a_out, b_out, a_in, b_in, bin_size;
  arg_format_t formatting;

  if (stbds_arrlen(positionals) != 5) {
    goto arg_error;
  }

  if (!ReadArguments(positionals[0], &a_out, "A outputs") ||
      !ReadArguments(positionals[1], &b_out, "B outputs") ||
      !ReadArguments(positionals[2], &a_in, "A inputs") ||
      !ReadArguments(positionals[3], &b_in, "B inputs") ||
      !ReadArguments(positionals[4], &bin_size, "bin size")) {
    goto arg_error;
  }

  formatting = kRowFmt;
  for (size_t i = 0; i < stbds_arrlenu(flags); i++) {
    char *flag = flags[i];
    if (strcmp(flag, "cg") == 0) {
      formatting = kCgFmt;
    } else if (strcmp(flag, "p") == 0) {
      formatting = kPFmt;
    } else if (strcmp(flag, "p-raw") == 0) {
      formatting = kPRawFmt;
    }
  }

  args_t result = {
      .a_in = a_in,
      .a_out = a_out,
      .b_in = b_in,
      .b_out = b_out,
      .bin_size = bin_size,
      .formatting = formatting,
  };

  stbds_arrfree(positionals);
  stbds_arrfree(flags);

  return result;

arg_error:
  stbds_arrfree(positionals);
  stbds_arrfree(flags);

  fprintf(stderr, "Usage:\n   ");
  if (argc > 0) {
    fprintf(stderr, "%s", argv[0]);
  } else {
    fprintf(stderr, "./main.exe");
  }
  fprintf(stderr,
          " [--cg|--p|--p-raw] <A outputs> <B outputs> <A inputs> <B inputs> "
          "<bin size>");
  exit(EXIT_FAILURE);
}

static _Bool RowsEquivalent(
    const size_t a_in, const size_t b_in, const size_t a_out,
    const size_t b_out, size_t row_len, data_t *p_buf, data_t *cg_buf,
    permutation_generator_t *a_in_perm, permutation_generator_t *b_in_perm,
    permutation_generator_t *a_out_perms, permutation_generator_t *b_out_perms,
    const data_t *restrict lhs, const data_t *restrict rhs) {
  PermutationReset(a_in_perm);
  PermutationReset(b_in_perm);
  for (size_t i = 0; i < a_in; i++) {
    PermutationReset(a_out_perms + i);
  }
  for (size_t i = 0; i < b_in; i++) {
    PermutationReset(b_out_perms + i);
  }

  while (!a_in_perm->exhausted) {
    PermutationReset(b_in_perm);
    while (!b_in_perm->exhausted) {
      ResetConditionalPermutations(a_out_perms, a_in);
      while (!a_out_perms[a_in - 1].exhausted) {
        ResetConditionalPermutations(b_out_perms, b_in);
        while (!b_out_perms[b_in - 1].exhausted) {
          // Compare the two rows

          FromCgToP(a_out, b_out, a_in, b_in, rhs, p_buf,
                    a_in_perm->permutation, b_in_perm->permutation, a_out_perms,
                    b_out_perms);

          {
            FromPToCg(a_out, b_out, a_in, b_in, cg_buf, p_buf);
            _Bool equivalent = 1;
            for (size_t i = row_len; i > 0; i--) {
              if (lhs[i - 1] != cg_buf[i - 1]) {
                equivalent = 0;
                break;
              }
            }

            if (equivalent) {
              return 1;
            }
          }

          // If the number of output labels are the same, and the number
          // of input labels is also the same, we can also check for
          // equality under party swapping. I don't expect this
          // conditional to be very penalizing because it's very
          // predictable.
          if (a_in == b_in && a_out == b_out) {
            // Use the results in p_buf
            PSwapParties(a_out, a_in, p_buf);
            FromPToCg(a_out, b_out, a_in, b_in, cg_buf, p_buf);
            _Bool equivalent = 1;
            for (size_t i = row_len; i > 0; i--) {
              if (lhs[i - 1] != cg_buf[i - 1]) {
                equivalent = 0;
                break;
              }
            }

            if (equivalent) {
              return 1;
            }
          }
          AdvanceConditionalPermutations(b_out_perms, b_in);
        }

        AdvanceConditionalPermutations(a_out_perms, a_in);
      }

      PermutationNext(b_in_perm);
    }

    PermutationNext(a_in_perm);
  }
  
  return 0;
}

typedef struct {
  // The updated pointer to the vector with the indices of the unique rows.
  size_t *unique;
  // The number of elements pushed to the vector.
  size_t count;
} reduction_result_t;

// Calculates inequivalent rows out of a given block of the rows. A block is a
// contiguous subset of rows, with indices
// `block_start..(block_start+row_count)`.
//
// Arguments:
//    a_in        Number of inputs for A
//    b_in        Number of inputs for B
//    a_out       Number of outputs for A
//    a_out       Number of outputs for B
//    matrix      The matrix whose rows are to be examined
//    row_len     Length of a matrix row
//    p_buf       data_t buffer capable of holding a P representation row
//    cg_buf      data_t buffer capable of holding a CG representation row
//    a_in_perm   permutation_generator_t for permutations of A's input
//    b_in_perm   permutation_generator_t for permutations of B's input
//    a_out_perms permutation_generator_t* array for permutations of A's output
//    b_out_perms permutation_generator_t* array for permutations of B's output
//    block_start Where to start counting the indices from.
//    row_count   Length of block
//    unique      size_t stbds vector pointer
//
// The vector `unique` is only pushed to; it's the user's responsibility to
// allocate or free it as needed.
//
// `row_count` MUST be greater than 0, otherwise behaviour is undefined.
//
// `p_buf`, and `cg_buf` WILL be clobbered. They don't need to be initialized,
// only allocated. It's the caller's responsibility to free these buffers if
// applicable.
//
// The permutation generators are expected to be correctly initialized; this
// function only resets the permutation generators, it does NOT allocate them
// or free them.
static reduction_result_t UniqueInBlock(
    const size_t a_in, const size_t b_in, const size_t a_out,
    const size_t b_out, const matrix_t *matrix, size_t row_len, data_t *p_buf,
    data_t *cg_buf, permutation_generator_t *a_in_perm,
    permutation_generator_t *b_in_perm, permutation_generator_t *a_out_perms,
    permutation_generator_t *b_out_perms, const size_t block_start,
    const size_t row_count, size_t *unique) {
  size_t *remaining = NULL;
  size_t *remaining_swap = NULL;
  size_t push_count = 0;

  // First round: perform the first split of equivalence classes.
  {
    stbds_arrput(unique, block_start);
    push_count++;
    data_t *root = matrix->head + block_start * matrix->row_len;

    for (size_t rhs_i = 1; rhs_i < row_count; rhs_i++) {
      data_t *rhs = matrix->head + (block_start + rhs_i) * matrix->row_len;
      _Bool equivalent = RowsEquivalent(a_in, b_in, a_out, b_out, row_len,
                                        p_buf, cg_buf, a_in_perm, b_in_perm,
                                        a_out_perms, b_out_perms, root, rhs);
      if (!equivalent) {
        stbds_arrput(remaining, rhs_i);
      }
    }
  }

  // Remaining rounds: perform reduction
  {
    while (stbds_arrlenu(remaining) > 0) {
      size_t lhs_i = stbds_arrpop(remaining);
      stbds_arrput(unique, block_start + lhs_i);
      push_count++;
      data_t *lhs = matrix->head + (block_start + lhs_i) * matrix->row_len;

      for (size_t i = 0; i < stbds_arrlenu(remaining); i++) {
        size_t rhs_i = remaining[i];
        data_t *rhs = matrix->head + (block_start + rhs_i) * matrix->row_len;

        _Bool equivalent = RowsEquivalent(a_in, b_in, a_out, b_out, row_len,
                                          p_buf, cg_buf, a_in_perm, b_in_perm,
                                          a_out_perms, b_out_perms, lhs, rhs);
        if (!equivalent) {
          stbds_arrput(remaining_swap, rhs_i);
        }
      }

      stbds_arrfree(remaining);
      remaining = remaining_swap;
      remaining_swap = NULL;
    }
  }

  stbds_arrfree(remaining);
  stbds_arrfree(remaining_swap);

  reduction_result_t result = {
      .unique = unique,
      .count = push_count,
  };

  return result;
}

// Calculates inequivalent rows out of a given subset of the rows. The subset
// is identified by a pair of arrays of indices.
//
// Arguments:
//    a_in              Number of inputs for A
//    b_in              Number of inputs for B
//    a_out             Number of outputs for A
//    a_out             Number of outputs for B
//    matrix            The matrix whose rows are to be examined
//    row_len           Length of a matrix row
//    p_buf             data_t buffer capable of holding a P representation row
//    cg_buf            data_t buffer capable of holding a CG representation row
//    a_in_perm         permutation_generator_t for permutations of A's input
//    b_in_perm         permutation_generator_t for permutations of B's input
//    a_out_perms       permutation_generator_t* array for permutations of A's
//                        output
//    b_out_perms       permutation_generator_t* array for permutations of B's
//                        output
//    first_row_idxs    First array of indices of rows to consider
//    first_row_count   Length of first_row_idxs[]
//    second_row_idxs   Second array of indices of rows to consider
//    second_row_count  Length of second_row_idxs[]
//    unique            size_t stbds vector pointer
//
// The vector `unique` is only pushed to; it's the user's responsibility to
// allocate or free it as needed.
//
// `first_row_count` and `second_row_count` MUST be greater than 0, otherwise
// behaviour is undefined.
//
// `seen`, `p_buf`, and `cg_buf` WILL be clobbered. They don't need to be
// initialized, only allocated. It's the caller's responsibility to free these
// buffers if applicable.
//
// The permutation generators are expected to be correctly initialized; this
// function only resets the permutation generators, it does NOT allocate them
// or free them.
static reduction_result_t UniqueInSubsetPair(
    const size_t a_in, const size_t b_in, const size_t a_out,
    const size_t b_out, const matrix_t *matrix, const size_t row_len,
    data_t *p_buf, data_t *cg_buf, permutation_generator_t *a_in_perm,
    permutation_generator_t *b_in_perm, permutation_generator_t *a_out_perms,
    permutation_generator_t *b_out_perms, const size_t *first_row_idxs,
    const size_t first_row_count, const size_t *second_row_idxs,
    const size_t second_row_count, size_t *unique) {
  // This one is a little more complicated:
  // Every two elements within `first_row_idxs` are inequivalent, by design, and
  //  likewise for `second_row_idxs`. This means that if a row A in `first_row`
  //  is equivalent to a row A' in `second_row`, it cannot be equivalent to any
  //  other row in `second_row` by transitivity of the equivalence.

  size_t push_count = 0;
  size_t *classes = NULL;

  for (size_t i = 0; i < first_row_count; i++) {
    size_t lhs_i = first_row_idxs[i];
    stbds_arrput(unique, lhs_i);
    push_count++;
    stbds_arrput(classes, lhs_i);
  }

  for (size_t i = 0; i < second_row_count; i++) {
    if (stbds_arrlenu(classes) == 0) {
      break;
    }

    size_t rhs_i = second_row_idxs[i];
    data_t *rhs = matrix->head + rhs_i * matrix->row_len;

    _Bool equivalent;

    for (size_t i = 0; i < stbds_arrlenu(classes); i++) {
      size_t lhs_i = classes[i];
      data_t *lhs = matrix->head + lhs_i * matrix->row_len;

      equivalent = RowsEquivalent(a_in, b_in, a_out, b_out, row_len, p_buf,
                                  cg_buf, a_in_perm, b_in_perm, a_out_perms,
                                  b_out_perms, lhs, rhs);

      if (equivalent) {
        stbds_arrdel(classes, i);
        break;
      }
    }

    if (!equivalent) {
      stbds_arrput(unique, rhs_i);
      push_count++;
    }
  }

  stbds_arrfree(classes);

  reduction_result_t result = {
      .unique = unique,
      .count = push_count,
  };

  return result;
}

int main(int argc, char *argv[]) {
  const args_t args = ParseArgs(argc, argv);
  const size_t a_out = args.a_out;
  const size_t b_out = args.b_out;
  const size_t a_in = args.a_in;
  const size_t b_in = args.b_in;
  const size_t bin_size = args.bin_size;

  if (a_out < 2 || b_out < 2 || a_in < 2 || b_in < 2) {
    fprintf(stderr,
            "This program is not prepared to deal with input or output label "
            "counts smaller than 2.");
    exit(EXIT_FAILURE);
  }

  const size_t row_len = (a_out - 1) * (b_out - 1) * a_in * b_in +
                         (a_out - 1) * a_in + (b_out - 1) * b_in + 1;
  matrix_t matrix = ParseMatrix(row_len);
  const size_t row_count = matrix.len / matrix.row_len;

  if (bin_size > row_count) {
    fprintf(stderr,
            "Bin size cannot be bigger than the number of rows in the input.");
    exit(EXIT_FAILURE);
  }

  // BEGINNING OF COMPARISON ALGORITHM

  const size_t bin_count =
      ((row_count % bin_size == 0) ? row_count / bin_size
                                   : row_count / bin_size + 1);

  size_t *unique = NULL;
  size_t *ends =
      NULL;  // Bin i contains elements with indices ends[i-1]..ends[i]

  pthread_rwlockattr_t attr;
  pthread_rwlockattr_init(&attr);
  pthread_rwlockattr_setpshared(&attr, PTHREAD_PROCESS_SHARED);
  pthread_rwlock_t vector_lock, vector_swap_lock;
  pthread_rwlock_init(&vector_lock, &attr);

#pragma omp parallel default(none)                                             \
    shared(stderr, args, a_out, b_out, a_in, b_in, row_len, matrix, row_count, \
               bin_size, bin_count, unique, ends, vector_lock)
  {
    data_t *p_buf = malloc(a_out * b_out * a_in * b_in * sizeof(data_t));
    data_t *cg_buf = malloc((((a_out - 1) * (b_out - 1) * a_in * b_in +
                              (a_out - 1) * a_in + (b_out - 1) * b_in) +
                             1) *
                            sizeof(data_t));

    if (p_buf == NULL) {
      fprintf(stderr, "Failed to allocate P notation buffer. Aborting.");
      exit(EXIT_FAILURE);
    }
    if (cg_buf == NULL) {
      fprintf(stderr, "Failed to allocate CG notation buffer. Aborting.");
      exit(EXIT_FAILURE);
    }

    permutation_generator_t a_in_perm = PermutationNewGenerator(a_in);
    permutation_generator_t b_in_perm = PermutationNewGenerator(b_in);
    permutation_generator_t *a_out_perms =
        malloc(a_in * sizeof(permutation_generator_t));
    permutation_generator_t *b_out_perms =
        malloc(b_in * sizeof(permutation_generator_t));

    if (a_out_perms == NULL || b_out_perms == NULL) {
      fprintf(
          stderr,
          "Failed to allocate space for conditioned permutation generators. "
          "Aborting.");
      exit(EXIT_FAILURE);
    }

    for (size_t i = 0; i < a_in; i++) {
      a_out_perms[i] = PermutationNewGenerator(a_out);
    }
    for (size_t i = 0; i < b_in; i++) {
      b_out_perms[i] = PermutationNewGenerator(b_out);
    }

    // Base round

    size_t *local_unique = NULL;
    size_t *local_ends = NULL;

#pragma omp for schedule(dynamic)
    for (size_t bin_i = 0; bin_i < bin_count; bin_i++) {
      const size_t block_i_start = bin_i * bin_size;
      size_t bin_i_size = row_count - block_i_start;
      if (bin_i_size > bin_size) {
        bin_i_size = bin_size;
      }

      reduction_result_t result =
          UniqueInBlock(a_in, b_in, a_out, b_out, &matrix, row_len, p_buf,
                        cg_buf, &a_in_perm, &b_in_perm, a_out_perms,
                        b_out_perms, block_i_start, bin_i_size, local_unique);

      local_unique = result.unique;
      size_t push_count = result.count;
      size_t ends_len = stbds_arrlenu(local_ends);
      if (ends_len == 0) {
        stbds_arrput(local_ends, push_count);
      } else {
        size_t last_end = stbds_arrlast(local_ends);
        stbds_arrput(local_ends, last_end + push_count);
      }

      fprintf(stderr, ".");
      fflush(stderr);

    }  // End of omp for
       // Implicit barrier

    // Merge locals
    pthread_rwlock_wrlock(&vector_lock);
    {
      for (size_t i = 0; i < stbds_arrlenu(local_unique); i++) {
        stbds_arrpush(unique, local_unique[i]);
      }
      size_t ends_base_value = 0;
      if (stbds_arrlen(ends) > 0) {
        ends_base_value = stbds_arrlast(ends);
      }
      for (size_t i = 0; i < stbds_arrlenu(local_ends); i++) {
        stbds_arrpush(ends, local_ends[i] + ends_base_value);
      }
    }
    pthread_rwlock_unlock(&vector_lock);

#pragma omp barrier

    stbds_arrfree(local_unique);
    stbds_arrfree(local_ends);
    local_unique = NULL;
    local_ends = NULL;

#pragma omp single
    {
      fprintf(stderr, "Base round complete\n");
      fflush(stderr);
    }

    // Reductions round

    size_t round_bin_count;
    pthread_rwlock_rdlock(&vector_lock);
    { round_bin_count = stbds_arrlenu(ends); }
    pthread_rwlock_unlock(&vector_lock);

    while (round_bin_count > 1) {
      const size_t pairs = round_bin_count / 2;

#pragma omp for schedule(dynamic)
      for (size_t pair_i = 0; pair_i < pairs; pair_i++) {
        size_t *first_block_idxs, *second_block_idxs;
        size_t first_block_len, second_block_len;

        if (pair_i == 0) {
          first_block_idxs = unique;
          first_block_len = ends[0];
        } else {
          first_block_idxs = unique + ends[2 * pair_i - 1];
          first_block_len = ends[2 * pair_i] - ends[2 * pair_i - 1];
        }
        second_block_idxs = unique + ends[2 * pair_i];
        second_block_len = ends[2 * pair_i + 1] - ends[2 * pair_i];

        reduction_result_t result = UniqueInSubsetPair(
            a_in, b_in, a_out, b_out, &matrix, row_len, p_buf, cg_buf,
            &a_in_perm, &b_in_perm, a_out_perms, b_out_perms, first_block_idxs,
            first_block_len, second_block_idxs, second_block_len, local_unique);

        local_unique = result.unique;
        size_t reduction_size = result.count;
        if (stbds_arrlenu(local_ends) == 0) {
          stbds_arrput(local_ends, reduction_size);
        } else {
          size_t last_end = stbds_arrlast(local_ends);
          stbds_arrput(local_ends, last_end + reduction_size);
        }

      }  // End of loop over pairs
         // Implicit barrier. Important because of the following #single

#pragma omp single
      {
        if (IsOdd(round_bin_count)) {
          // Push the remaining bin into the swap variables
          size_t *odd_block_idxs = unique + ends[stbds_arrlenu(ends) - 2];
          size_t odd_block_len =
              stbds_arrlast(ends) - ends[stbds_arrlenu(ends) - 2];

          for (size_t i = 0; i < odd_block_len; i++) {
            stbds_arrput(local_unique, odd_block_idxs[i]);
          }
          if (stbds_arrlenu(local_ends) > 0) {
            size_t last_end = stbds_arrlast(local_ends);
            stbds_arrput(local_ends, last_end + odd_block_len);
          } else {
            stbds_arrput(local_ends, odd_block_len);
          }
        }

        stbds_arrfree(unique);
        stbds_arrfree(ends);
        unique = NULL;
        ends = NULL;
      }  // End of single section
         // Impilcit barrier. Important due to the clearing of unique, local.

      // Merge locals into global
      pthread_rwlock_wrlock(&vector_lock);
      {
        for (size_t i = 0; i < stbds_arrlenu(local_unique); i++) {
          stbds_arrpush(unique, local_unique[i]);
        }
        size_t ends_base_value = 0;
        if (stbds_arrlen(ends) > 0) {
          ends_base_value = stbds_arrlast(ends);
        }
        for (size_t i = 0; i < stbds_arrlenu(local_ends); i++) {
          stbds_arrpush(ends, local_ends[i] + ends_base_value);
        }
      }
      pthread_rwlock_unlock(&vector_lock);

      stbds_arrfree(local_unique);
      stbds_arrfree(local_ends);
      local_unique = NULL;
      local_ends = NULL;

#pragma omp barrier

      pthread_rwlock_rdlock(&vector_lock);
      { round_bin_count = stbds_arrlenu(ends); }
      pthread_rwlock_unlock(&vector_lock);
    }  // End of pairs reduction loop

    // Program is done!
    // `unique` should hold the indices of the unique rows.

#pragma omp single
    {
      if (args.formatting == kPFmt) {
        PermutationReset(&a_in_perm);
        PermutationReset(&b_in_perm);
        for (size_t i = 0; i < a_in; i++) {
          PermutationReset(a_out_perms + i);
        }
        for (size_t i = 0; i < b_in; i++) {
          PermutationReset(b_out_perms + i);
        }
      }

      // Print every unique row
      size_t unique_count = stbds_arrlenu(unique);
      for (size_t i = 0; i < unique_count; i++) {
        size_t row_idx = unique[i];
        switch (args.formatting) {
          default:
          case kRowFmt: {
            PrintMatrixRow(&matrix, row_idx);
          } break;
          case kCgFmt: {
            printf("%zu: ", i);
            PrintCg(a_out, b_out, a_in, b_in,
                    matrix.head + row_idx * matrix.row_len);
          } break;
          case kPFmt: {
            printf("%zu: ", i);
            FromCgToP(a_out, b_out, a_in, b_in,
                      matrix.head + row_idx * matrix.row_len, p_buf,
                      a_in_perm.permutation, b_in_perm.permutation, a_out_perms,
                      b_out_perms);
            PrintP(a_out, b_out, a_in, b_in, p_buf);
          } break;
          case kPRawFmt: {
            FromCgToP(a_out, b_out, a_in, b_in,
                      matrix.head + row_idx * matrix.row_len, p_buf,
                      a_in_perm.permutation, b_in_perm.permutation, a_out_perms,
                      b_out_perms);
            PrintPRaw(a_out, b_out, a_in, b_in, p_buf);
          } break;
        }
      }
    }  // End of omp single region

    PermutationFree(&a_in_perm);
    PermutationFree(&b_in_perm);
    for (size_t i = 0; i < a_in; i++) {
      PermutationFree(a_out_perms + i);
    }
    free(a_out_perms);
    for (size_t i = 0; i < b_in; i++) {
      PermutationFree(b_out_perms + i);
    }
    free(b_out_perms);
    free(cg_buf);
    free(p_buf);
  }  // End of omp parallel region

  // Free all the memory so the sanitizer is happy.
  pthread_rwlock_destroy(&vector_lock);
  pthread_rwlock_destroy(&vector_swap_lock);
  stbds_arrfree(unique);
  stbds_arrfree(ends);
  MatrixFree(&matrix);
  return EXIT_SUCCESS;
}
